//  Copyright 2019 Kakao Corp.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import UIKit

// MARK: - codable helper
public enum AnyValue : Codable {
    case string(String)
    case int(Int)
    case bool(Bool)
    
    public init(from decoder : Decoder) throws {
        let container = try decoder.singleValueContainer()
        if let int = try? container.decode(Int.self) {
            self = .int(int)
            return
        }
        if let string = try? container.decode(String.self) {
            self = .string(string)
            return
        }
        if let bool = try? container.decode(Bool.self) {
            self = .bool(bool)
            return
        }
        throw AnyValueError.missingValue
    }
    
    public func encode(to encoder: Encoder) throws {
        var container = encoder.singleValueContainer()
        switch self {
        case .string(let value):
            try container.encode(value)
        case .int(let value):
            try container.encode(value)
        case .bool(let value):
            try container.encode(value)
        }
    }
    
    enum AnyValueError:Error {
            case missingValue
        }
}

public struct ApiParam : Codable {
    public let key: String
    public let value: AnyValue
}

public struct ApiMetaInfo : Codable {
    public let name: String
    public let params: [ApiParam]?
}

//MARK: - Model

public enum ParamValue {
    case editableText(String)
    case multiSelectableItem([String])
    case singleSelectableItem([String])
}

public class ParamItem {
    public var key: String
    public var value: ParamValue
    public var selected: Bool
    public var selectedIndex:[Int]
    
    public init(key: String, value: ParamValue, selectedIndex: [Int] = [], selected: Bool = false) {
        self.key = key
        self.value = value
        self.selectedIndex = selectedIndex
        self.selected = selected
    }
    
    public func update(newValue:String? = nil) {
        switch value {
        case .editableText(_):
            if let newValue = newValue {
                value = .editableText(newValue)
            }
        default:
            break
        }
    }
    
    public func select() {
        switch value {
        case .editableText(_):
            selected = !selected
        default:
            break
        }
    }
    
    public func select(index:Int) -> Bool {
        var selectedAtIndex = false
        switch value {
        case .multiSelectableItem(_):
            if selectedIndex.contains(index) {
                selectedIndex = selectedIndex.filter({ oldIndex in
                    return oldIndex != index
                })
                selectedAtIndex = false
            }
            else {
                selectedIndex.append(index)
                selectedAtIndex = true
            }
            (selectedIndex.count > 0) ? (selected = true) : (selected = false)
        case .singleSelectableItem(_):
            if selectedIndex.contains(index) {
                selectedIndex.removeAll()
                selectedAtIndex = false
            }
            else {
                selectedIndex.removeAll()
                selectedIndex.append(index)
                selectedAtIndex = true
            }
            (selectedIndex.count > 0) ? (selected = true) : (selected = false)
        default:
            break
        }
        return selectedAtIndex
    }
    
    public func isSelectedAt(index:Int) -> Bool {
        var isSelected = false
        
        switch value {
        case .multiSelectableItem(_):
            fallthrough
        case .singleSelectableItem(_):
            if selectedIndex.contains(index) {
                isSelected = true
            }
        default:
            break
        }
        return isSelected
    }
}
public class Api {
    public var name: String
    public var paramItems: [ParamItem]
    public init(name:String, paramItems:[ParamItem]) {
        self.name = name
        self.paramItems = paramItems
    }
}

class CommonConstants {
    static let paramKeyAreaWidth: CGFloat = 80
    static let paramValueAreaWidth: CGFloat = 200
    static let paramValueAreaHeight: CGFloat = 40
    static let paramUIRadius: CGFloat = 5
    static let paramCellSpacing: CGFloat = 8
    static let paramCollectionViewCellFontSize: CGFloat = 14
}

//MARK: - SelectableCollectionViewCell
class SelectableCollectionViewCell: UICollectionViewCell {
    var titleLabel: UILabel = {
        let label = UILabel(frame: CGRect.zero)

        return label
    }()
    
    override init(frame: CGRect) {
        super.init(frame: frame)
        titleLabel.font = UIFont.systemFont(ofSize: CommonConstants.paramCollectionViewCellFontSize, weight: .regular)
        titleLabel.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
        titleLabel.textAlignment = .center
        titleLabel.textColor = UIColor.black
        titleLabel.adjustsFontSizeToFitWidth = true
        
        contentView.backgroundColor = UIColor.init(red: 0.1, green: 0.1, blue: 0.1, alpha: 0.1)
        contentView.addSubview(titleLabel)
        
        addSubviewsConstraint()
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    func addSubviewsConstraint() {
        print("MultiCheckBoxCollectionViewCell addSubviewsConstraint() called!")
        
        titleLabel.translatesAutoresizingMaskIntoConstraints = false
        titleLabel.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        titleLabel.heightAnchor.constraint(equalTo: contentView.heightAnchor).isActive = true
        titleLabel.widthAnchor.constraint(equalTo: contentView.widthAnchor).isActive = true
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
        
//        self.titleLabel.frame = CGRect(x:0, y:0, width:self.frame.width, height:self.frame.height)
    }

    override func prepareForReuse() {
        super.prepareForReuse()
    }
    
    public func updateCollectionViewCellUI(selected:Bool) {
        setSelected(selected)
    }
    
    public func setSelected(_ cellSelected:Bool) {
        if cellSelected {
            print("selected")
            titleLabel.textColor = UIColor.white
            self.contentView.backgroundColor = UIColor.init(red: 0.1, green: 0.1, blue: 0.1, alpha: 0.9)
        } else {
            print("not selected")
            titleLabel.textColor = UIColor.black
            self.contentView.backgroundColor = UIColor.init(red: 0.1, green: 0.1, blue: 0.1, alpha: 0.1)
        }
    }
}

enum SelectMode {
    case single
    case multi
}

//MARK: - SelectParamTableViewCell
class SelectParamTableViewCell: ParamTableViewCell {
    var selectMode: SelectMode = .single //default
    var multipleCheckBoxCollectionView: UICollectionView = {
        let layout = UICollectionViewFlowLayout()
        layout.minimumInteritemSpacing = 2
        layout.minimumLineSpacing = 2
        layout.scrollDirection = .horizontal
        let collectionView = UICollectionView(frame: CGRect.zero, collectionViewLayout: layout)
        collectionView.backgroundColor = UIColor.white
        return collectionView
    }()
    
    override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
        super.init(style: style, reuseIdentifier: reuseIdentifier)
        
        multipleCheckBoxCollectionView.delegate = self
        multipleCheckBoxCollectionView.dataSource = self
        multipleCheckBoxCollectionView.showsHorizontalScrollIndicator = false
        multipleCheckBoxCollectionView.register(SelectableCollectionViewCell.self, forCellWithReuseIdentifier: "SelectableCollectionViewCell")
        contentView.addSubview(multipleCheckBoxCollectionView)
        
        addSubviewsConstraint()
    }

    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    override func addSubviewsConstraint() {
        print("MultipleSelectParamTableViewCell addSubviewsConstraint() called!")
        super.addSubviewsConstraint()
        multipleCheckBoxCollectionView.translatesAutoresizingMaskIntoConstraints = false
        multipleCheckBoxCollectionView.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        multipleCheckBoxCollectionView.leadingAnchor.constraint(equalTo: paramKeyLabel.trailingAnchor, constant: 4).isActive = true
        multipleCheckBoxCollectionView.trailingAnchor.constraint(equalTo: contentView.trailingAnchor, constant: -4).isActive = true
        multipleCheckBoxCollectionView.heightAnchor.constraint(equalTo: contentView.heightAnchor).isActive = true
    }

    override func layoutSubviews() {
        super.layoutSubviews()
    }

    override func prepareForReuse() {
        super.prepareForReuse()
        
        multipleCheckBoxCollectionView.delegate = nil
        multipleCheckBoxCollectionView.dataSource = nil
    }
    
    public func paramItemValues(_ paramItem:ParamItem?) -> [String]? {
        guard let paramItem = paramItem else { return [] }
        
        var apiParamValues : [String]? = nil
        switch paramItem.value {
        case .multiSelectableItem(let values):
            apiParamValues = values
        case .singleSelectableItem(let values):
            apiParamValues = values
        default:
            break
        }
        return apiParamValues
    }
    
    override public func setParamItem(_ paramItem:ParamItem) {
        super.setParamItem(paramItem)
        setNeedsLayout()
    }
    
    public func selectItem(index:Int) {
        paramItem?.select(index:index)
//        updateSwitchUI(paramItem?.selected ?? false)
    }
}

extension SelectParamTableViewCell: UICollectionViewDelegate, UICollectionViewDataSource {
    func collectionView(_ collectionView: UICollectionView, numberOfItemsInSection section: Int) -> Int {
        return paramItemValues(self.paramItem)?.count ?? 0
    }
    
    func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
        guard let paramValues = paramItemValues(self.paramItem) else {
            return SelectableCollectionViewCell()
        }
        
        if let cell = collectionView.dequeueReusableCell(withReuseIdentifier: "SelectableCollectionViewCell", for: indexPath) as? SelectableCollectionViewCell {
            cell.titleLabel.text = paramValues[indexPath.item]
            cell.backgroundColor = .white
            cell.layer.cornerRadius = CommonConstants.paramUIRadius
            cell.layer.masksToBounds = true
            return cell
        }
        return SelectableCollectionViewCell()
    }
    
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        if let collectionViewCell = collectionView.cellForItem(at: indexPath) as? SelectableCollectionViewCell {
            //model update
            guard let paramItem = self.paramItem else { return }
            let selectedAtIndex = paramItem.select(index: indexPath.item)
            
            //UI update
            updateSwitchUI()
            if selectMode == .single {
                resetCollectionViewCellUI()
            }
            collectionViewCell.updateCollectionViewCellUI(selected:selectedAtIndex)
        }
    }
    
    func resetCollectionViewCellUI() {
        if let visibleCells = self.multipleCheckBoxCollectionView.visibleCells as? [SelectableCollectionViewCell] {
            visibleCells.forEach { cell in
                cell.setSelected(false)
            }
        }
    }
}

extension String {
    func widthOfString(usingFont font: UIFont) -> CGFloat {
        let fontAttributes = [NSAttributedString.Key.font: font]
        let size = self.size(withAttributes: fontAttributes)
        return size.width
    }

    func heightOfString(usingFont font: UIFont) -> CGFloat {
        let fontAttributes = [NSAttributedString.Key.font: font]
        let size = self.size(withAttributes: fontAttributes)
        return size.height
    }

    func sizeOfString(usingFont font: UIFont) -> CGSize {
        let fontAttributes = [NSAttributedString.Key.font: font]
        return self.size(withAttributes: fontAttributes)
    }
}

extension SelectParamTableViewCell: UICollectionViewDelegateFlowLayout {
    func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, insetForSectionAt section: Int) -> UIEdgeInsets {
            return UIEdgeInsets(top: 0, left: CommonConstants.paramCellSpacing, bottom: 0, right: CommonConstants.paramCellSpacing)
        }
    
    func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, sizeForItemAt indexPath: IndexPath) -> CGSize {
        guard let paramValues = paramItemValues(self.paramItem) else { return CGSize(width: 100, height: 40) }
        let valueText : String = paramValues[indexPath.item]
        let width = valueText.widthOfString(usingFont:UIFont.systemFont(ofSize: CommonConstants.paramCollectionViewCellFontSize, weight: .regular))
        return CGSize(width: width + (CommonConstants.paramUIRadius * 2), height: 40)
    }
}

//MARK: - EditableTextParamTableViewCell
extension UITextField {
    func setLeftPaddingPoints(_ amount:CGFloat){
        let paddingView = UIView(frame: CGRect(x: 0, y: 0, width: amount, height: self.frame.size.height))
        self.leftView = paddingView
        self.leftViewMode = .always
    }
    func setRightPaddingPoints(_ amount:CGFloat) {
        let paddingView = UIView(frame: CGRect(x: 0, y: 0, width: amount, height: self.frame.size.height))
        self.rightView = paddingView
        self.rightViewMode = .always
    }
}

class EditableTextParamTableViewCell : ParamTableViewCell, UITextFieldDelegate {
    var paramValueTextField: UITextField = {
        let textField = UITextField(frame: CGRect.zero)
        textField.setLeftPaddingPoints(CommonConstants.paramCellSpacing)
        textField.setRightPaddingPoints(CommonConstants.paramCellSpacing)
        textField.backgroundColor = UIColor.init(red: 0.1, green: 0.1, blue: 0.1, alpha: 0.1)
        textField.layer.cornerRadius = CommonConstants.paramUIRadius
        textField.layer.masksToBounds = true
        return textField
    }()
    
    override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
        super.init(style: style, reuseIdentifier: reuseIdentifier)
        paramValueTextField.delegate = self
        self.contentView.addSubview(self.paramValueTextField)
        
        addSubviewsConstraint()
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    override func addSubviewsConstraint() {
        print("EditableTextParamTableViewCell addSubviewsConstraint() called!")
        super.addSubviewsConstraint()
        paramValueTextField.translatesAutoresizingMaskIntoConstraints = false
        paramValueTextField.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        paramValueTextField.leadingAnchor.constraint(equalTo: paramKeyLabel.trailingAnchor, constant: 4).isActive = true
        paramValueTextField.trailingAnchor.constraint(equalTo: contentView.trailingAnchor, constant: -4).isActive = true
        paramValueTextField.heightAnchor.constraint(equalToConstant: CommonConstants.paramValueAreaHeight).isActive = true
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
//        self.paramValueTextField.frame = CGRect(x:self.paramKeyLabel.frame.size.width + self.gap,
//                                                y:0,
//                                                width:paramValueWidth,
//                                                height:paramCellHeight)
    }
    
    override func prepareForReuse() {
        super.prepareForReuse()
        self.paramValueTextField.text = nil
        paramValueTextField.delegate = nil
    }
    
    override public func setParamItem(_ paramItem:ParamItem) {
        super.setParamItem(paramItem)
        
        if case .editableText(let defaultValue) = paramItem.value {
            self.paramValueTextField.text = defaultValue
        }
    }
    
    public func selectItem() {
        paramItem?.select()
//        updateSwitchUI(paramItem?.selected ?? false)
    }
    
    func updateItem(text:String) {
        paramItem?.update(newValue: text)
        //self.selectItem() //값 바꾸는것과 선택은 별개
    }
    
    // UITextField Delegates
    func textFieldDidBeginEditing(_ textField: UITextField) {
    }
    func textFieldDidEndEditing(_ textField: UITextField) {
        updateItem(text:textField.text!)
    }
    func textFieldShouldBeginEditing(_ textField: UITextField) -> Bool {
        return true;
    }
    func textFieldShouldClear(_ textField: UITextField) -> Bool {
        return true;
    }
    func textFieldShouldEndEditing(_ textField: UITextField) -> Bool {
        return true;
    }
    func textField(_ textField: UITextField, shouldChangeCharactersIn range: NSRange, replacementString string: String) -> Bool {
        return true;
    }
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        textField.resignFirstResponder();
        return true;
    }
}



//MARK: - ParamCell
class ParamTableViewCell: UITableViewCell {

    var paramItem: ParamItem?
    
    var useParamSwitch: UISwitch = {
        let paramSwitch = UISwitch()
        return paramSwitch
    }()
    
    var paramKeyLabel: UILabel = {
        let label = UILabel(frame: CGRect.zero)
        return label
    }()
    
    override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
        super.init(style: style, reuseIdentifier: reuseIdentifier)
        
//        self.useParamSwitch.backgroundColor = UIColor.blue
        self.useParamSwitch.isOn = false
        self.useParamSwitch.isUserInteractionEnabled = false
        self.contentView.addSubview(self.useParamSwitch)
        
        self.paramKeyLabel.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
        self.paramKeyLabel.font = UIFont.boldSystemFont(ofSize: 18)
        self.paramKeyLabel.textColor = UIColor.black
        self.paramKeyLabel.textAlignment = .right
        self.paramKeyLabel.adjustsFontSizeToFitWidth = true
        self.contentView.addSubview(self.paramKeyLabel)
        
        self.accessoryType = .none
        self.contentView.backgroundColor = UIColor.white
        
//        addSubviewsConstraint()
    }
    
    func addSubviewsConstraint() {
        print("ParamTableViewCell addSubviewsConstraint() called!")
        useParamSwitch.translatesAutoresizingMaskIntoConstraints = false
        useParamSwitch.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        useParamSwitch.leadingAnchor.constraint(equalTo: contentView.leadingAnchor, constant: CommonConstants.paramCellSpacing).isActive = true
        
        paramKeyLabel.translatesAutoresizingMaskIntoConstraints = false
        paramKeyLabel.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        paramKeyLabel.leadingAnchor.constraint(equalTo: useParamSwitch.trailingAnchor, constant: CommonConstants.paramCellSpacing).isActive = true
        paramKeyLabel.widthAnchor.constraint(equalToConstant: CommonConstants.paramKeyAreaWidth).isActive = true
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
    }
    
    override func prepareForReuse() {
        super.prepareForReuse()
        
        self.useParamSwitch.isOn = false
        self.paramKeyLabel.text = nil
        self.accessoryType = .none
    }
    
    public func setParamItem(_ paramItem:ParamItem) {
        self.paramItem = paramItem
        self.paramKeyLabel.text = paramItem.key
    }
    
    public func updateSwitchUI() {
        if let paramItem = self.paramItem {
            useParamSwitch.setOn(paramItem.selected, animated: true)
        }
    }
}

class ParamNaviView: UIView {
    var leftButton: UIButton!
    var rightButton: UIButton!
    var titleLabel: UILabel!
    
    var titleLabelWidthConstraint: NSLayoutConstraint!
        
    override init(frame: CGRect) {
        super.init(frame: frame)
        
        self.backgroundColor = backgroundColor
        
//        leftButton = UIButton(frame: CGRect.zero)
//        leftButton.titleLabel?.font = UIFont.systemFont(ofSize: 14, weight: .semibold)
//        leftButton.setTitle("X", for: .normal)
////        leftButton.backgroundColor = UIColor.blue
//        leftButton.setTitleColor(UIColor.black, for: .normal)
//        leftButton.setTitleColor(UIColor.gray, for: .highlighted)
//        addSubview(leftButton)
        
        rightButton = UIButton(frame: CGRect.zero)
        rightButton.titleLabel?.font = UIFont.systemFont(ofSize: 18, weight: .semibold)
//        rightButton.backgroundColor = UIColor.blue
        rightButton.setTitle("X", for: .normal)
        rightButton.setTitleColor(UIColor.black, for: .normal)
        rightButton.setTitleColor(UIColor.gray, for: .highlighted)
        addSubview(rightButton)
        
        titleLabel = UILabel(frame: .zero)
        titleLabel.font = UIFont.systemFont(ofSize: 18, weight: .semibold)
        titleLabel.textAlignment = .center
        titleLabel.numberOfLines = 1
        titleLabel.backgroundColor = UIColor.white
        addSubview(titleLabel)
        
//        self.backgroundColor = UIColor.red
        
        addSubviewsConstraint()
    }
    
    func addSubviewsConstraint() {
//        leftButton.translatesAutoresizingMaskIntoConstraints = false
//        leftButton.centerYAnchor.constraint(equalTo:centerYAnchor).isActive = true
//        leftButton.leadingAnchor.constraint(equalTo: leadingAnchor, constant: 16).isActive = true
//        leftButton.widthAnchor.constraint(equalToConstant: 24).isActive = true
//        leftButton.heightAnchor.constraint(equalToConstant: 24).isActive = true
        
        titleLabel.translatesAutoresizingMaskIntoConstraints = false
        titleLabel.centerXAnchor.constraint(equalTo:centerXAnchor).isActive = true
        titleLabel.centerYAnchor.constraint(equalTo:centerYAnchor).isActive = true
        titleLabelWidthConstraint = titleLabel.widthAnchor.constraint(equalToConstant: 100)
        titleLabelWidthConstraint.isActive = true
        titleLabel.heightAnchor.constraint(equalToConstant: 21.5).isActive = true
        
        rightButton.translatesAutoresizingMaskIntoConstraints = false
        rightButton.centerYAnchor.constraint(equalTo:centerYAnchor).isActive = true
        rightButton.trailingAnchor.constraint(equalTo: trailingAnchor, constant: -16).isActive = true
        rightButton.widthAnchor.constraint(equalToConstant: 24).isActive = true
        rightButton.heightAnchor.constraint(equalToConstant: 24).isActive = true
    }
    
    func updateSubviewsConstraint() {
        if let text = titleLabel.text {
            let viewWidth = frame.width
            let maxWidth = ((viewWidth / 2) - rightButton.frame.width + 16) * 2 //중간부터 시작해서 오른쪽 카운트레이블과 4pt 간격을 둔 최대 길이 * 2

            titleLabelWidthConstraint.isActive = false
            titleLabelWidthConstraint.constant = getTextLabelWidth(text: text, font: titleLabel.font, maxWidth: maxWidth)
            titleLabelWidthConstraint.isActive = true
        }
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
        updateSubviewsConstraint()
    }
    
    func setTitle(title: String) {
        titleLabel.text = title
        setNeedsLayout()
    }
    
    func setRightButtonTitle(title: String) {
        rightButton.setTitle(title, for: .normal)
    }
    
    func setLeftButtonAction(target:Any?, action:Selector) {
        leftButton.addTarget(target, action: action, for: .touchUpInside)
    }
    
    func setRightButtonAction(target:Any?, action:Selector) {
        rightButton.addTarget(target, action: action, for: .touchUpInside)
    }
}

extension String {
    func getLabelWidth(usedFont: UIFont, maxWidth:CGFloat? = nil) -> CGFloat {
        let width : CGFloat = self.size(withAttributes: [NSAttributedString.Key.font : usedFont]).width
        if let maxWidth = maxWidth {
            return (width >= maxWidth) ? maxWidth : width
        }
        else {
            return width
        }
    }
}

extension ParamNaviView {
    func getTextLabelWidth(text:String? = nil, font:UIFont? = nil, maxWidth:CGFloat? = nil) -> CGFloat {
        if let font = font, let width = text?.getLabelWidth(usedFont: font, maxWidth: maxWidth) {
            return width + 2
        }
        else {
            return 0
        }
    }
}

//MARK: - ParamViewController
public class SDKSampleParamViewController: UIViewController, UITableViewDelegate, UITableViewDataSource {
    var backgroundView: UIView!
    var backgroudViewWidthConstraint : NSLayoutConstraint!
    var backgroudViewHeightConstraint : NSLayoutConstraint!
    
    let naviView: ParamNaviView = {
        let naviView = ParamNaviView(frame: CGRect.zero)
        return naviView
    }()
    let tableView: UITableView = {
        let tableView = UITableView(frame: CGRect.zero, style: .plain)
        return tableView
    }()
    let sendButton: UIButton = {
        let button = UIButton(frame: CGRect.zero)
        return button
    }()
    
    var safeAreaLayoutGuide : UILayoutGuide!
    var sections: [SectionOfRow]?
    
    var apiInfo: Api?
    var paramViewCompletionHandler: (([ParamItem]) -> Void)?
    var viewTitle: String?
    var paramItems: [ParamItem]?
    
    private let maxWidth: CGFloat = 360
    private let maxHeight: CGFloat = 640
    
    public convenience init(api:Api,
                            paramViewCompletionHandler: (([ParamItem]) -> Void)? = nil) {
        self.init()
        self.apiInfo = api
        self.viewTitle = api.name
        self.paramItems = api.paramItems
        self.paramViewCompletionHandler = paramViewCompletionHandler
    }
    
    public init() {
        super.init(nibName: nil, bundle: nil)
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    override public func loadView() {
        super.loadView()
        self.safeAreaLayoutGuide = view.layoutMarginsGuide
        
        setupViews()
    }
    
    override public func viewDidLoad() {
        super.viewDidLoad()
        
//        if #available(iOS 15.0, *) {
//            let navBarAppearance = UINavigationBarAppearance()
//            navBarAppearance.configureWithOpaqueBackground()
//            //Configure additional customizations here
//            UINavigationBar.appearance().standardAppearance = navBarAppearance
//            UINavigationBar.appearance().scrollEdgeAppearance = navBarAppearance
//        }
        
        extendedLayoutIncludesOpaqueBars = true
        self.tableView.delegate = self
        self.tableView.dataSource = self
        self.naviView.setTitle(title: "\(self.viewTitle ?? "")")
    }
    
    public func setupViews() {
        backgroundView = UIView(frame: view.frame)
        backgroundView.backgroundColor = UIColor.white
        backgroundView.layer.cornerRadius = CommonConstants.paramUIRadius
        
//        naviView.setLeftButtonAction(target: self, action: #selector(closeButtonAction))
        naviView.setRightButtonAction(target: self, action: #selector(closeButtonAction))
        backgroundView.addSubview(naviView)
        
        tableView.backgroundColor = .lightGray
        tableView.separatorStyle = UITableViewCell.SeparatorStyle.none
        backgroundView.addSubview(tableView)
        if #available(iOS 15, *) {
            tableView.sectionHeaderTopPadding = 0
        }
        tableView.allowsMultipleSelection = true
        tableView.register(EditableTextParamTableViewCell.self, forCellReuseIdentifier: "EditableTextParamTableViewCell")
        tableView.register(SelectParamTableViewCell.self, forCellReuseIdentifier: "SelectParamTableViewCell")
        
        sendButton.titleLabel?.font = UIFont.systemFont(ofSize: 18, weight: .semibold)
        sendButton.backgroundColor = UIColor.yellow
        sendButton.setTitle("Send", for: .normal)
        sendButton.setTitleColor(UIColor.black, for: .normal)
        sendButton.setTitleColor(UIColor.gray, for: .highlighted)
        
        sendButton.layer.cornerRadius = CommonConstants.paramUIRadius
        sendButton.layer.masksToBounds = true
        sendButton.layer.maskedCorners = [.layerMinXMaxYCorner, .layerMaxXMaxYCorner]
        sendButton.addTarget(self, action: #selector(okButtonAction), for: .touchUpInside)
        
        backgroundView.addSubview(sendButton)
        
        view.backgroundColor = UIColor.black.withAlphaComponent(0.7)
        view.addSubview(backgroundView)
        
        addSubviewsConstraint()
    }
    
    public func addSubviewsConstraint() {
        let size = calculateBackgroundSize()
        backgroundView.translatesAutoresizingMaskIntoConstraints = false
        backgroudViewWidthConstraint = backgroundView.widthAnchor.constraint(equalToConstant: size.width)
        backgroudViewWidthConstraint.isActive = true
        backgroudViewHeightConstraint = backgroundView.heightAnchor.constraint(equalToConstant: size.height)
        backgroudViewHeightConstraint.isActive = true

        backgroundView.centerXAnchor.constraint(equalTo:view.safeAreaLayoutGuide.centerXAnchor).isActive = true
        backgroundView.centerYAnchor.constraint(equalTo:view.safeAreaLayoutGuide.centerYAnchor).isActive = true
        
        naviView.translatesAutoresizingMaskIntoConstraints = false
        naviView.topAnchor.constraint(equalTo:backgroundView.topAnchor, constant: CommonConstants.paramUIRadius).isActive = true
        naviView.leadingAnchor.constraint(equalTo: backgroundView.leadingAnchor).isActive = true
        naviView.trailingAnchor.constraint(equalTo: backgroundView.trailingAnchor).isActive = true
        naviView.heightAnchor.constraint(equalToConstant: getNaviViewHeight()).isActive = true
        
        tableView.translatesAutoresizingMaskIntoConstraints = false
        tableView.topAnchor.constraint(equalTo: naviView.bottomAnchor).isActive = true
        tableView.leadingAnchor.constraint(equalTo: backgroundView.leadingAnchor).isActive = true
        tableView.trailingAnchor.constraint(equalTo: backgroundView.trailingAnchor).isActive = true
        tableView.heightAnchor.constraint(equalToConstant: size.height - 40 - getNaviViewHeight() - CommonConstants.paramUIRadius).isActive = true
//        tableView.bottomAnchor.constraint(equalTo: sendButton.topAnchor).isActive = true
        
        sendButton.translatesAutoresizingMaskIntoConstraints = false
        sendButton.topAnchor.constraint(equalTo: tableView.bottomAnchor).isActive = true
        sendButton.leadingAnchor.constraint(equalTo: backgroundView.leadingAnchor).isActive = true
        sendButton.trailingAnchor.constraint(equalTo: backgroundView.trailingAnchor).isActive = true
        sendButton.heightAnchor.constraint(equalToConstant: 40).isActive = true
    }
    
    @objc func okButtonAction(sender: Any?) {
        if let completionHandler = paramViewCompletionHandler {
            
            let selectedParamItems = apiInfo?.paramItems.filter({ paramItem in
                return (paramItem.selected == true)
            })
            completionHandler(selectedParamItems ?? [])
            self.dismiss(animated: true)
//            self.navigationController?.popViewController(animated: true)
        }
    }
    
    @objc func closeButtonAction(sender: Any?) {
        dismiss(animated: true) { [unowned self] in
        }
    }
    
    public func numberOfSections(in tableView: UITableView) -> Int {
        return 1
    }
    
    func getNaviViewHeight() -> CGFloat {
        var height: CGFloat = 44
        if (UIDevice.current.orientation.isLandscape) {
            height = 32
        }
        return height
    }
    
    func getPopupMargine() -> CGFloat {
        var margin  : CGFloat = 30
        if (UIDevice.current.orientation.isLandscape) {
            margin = 20
        }
        return margin
    }
    
    func calculateBackgroundSize() -> CGSize {
        var safeAreaLayoutFrame : CGRect!
        if let root = UIApplication.getMostTopViewController() {
            if #available(iOS 11, *) {
                safeAreaLayoutFrame = root.view.safeAreaLayoutGuide.layoutFrame
            }
            else {
                safeAreaLayoutFrame = view.bounds
            }
        }
        else {
            safeAreaLayoutFrame = view.bounds
        }
        let popupMargine = getPopupMargine()
        
        let width = (safeAreaLayoutFrame.width - (popupMargine * 2)) > maxWidth ? maxWidth : safeAreaLayoutFrame.width - (popupMargine * 2)
        let height = (safeAreaLayoutFrame.height - (popupMargine * 2)) > maxHeight ? maxHeight : safeAreaLayoutFrame.height - (popupMargine * 2)
        
        print("view width:\(self.view.frame.width) height:\(self.view.frame.height)")
        print("width:\(width) height:\(height)")
        
        return CGSize(width: width, height: height)
    }
    

    public func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if let items = self.paramItems {
            return items.count
        }
        else {
            return 0
        }
    }
    
    public func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let paramItem = self.paramItems![indexPath.row]
        var cell : ParamTableViewCell
        switch paramItem.value {
        case .editableText(_):
            cell = tableView.dequeueReusableCell(withIdentifier: "EditableTextParamTableViewCell", for: indexPath) as! EditableTextParamTableViewCell
        case .multiSelectableItem(_):
            cell = tableView.dequeueReusableCell(withIdentifier: "SelectParamTableViewCell", for: indexPath) as! SelectParamTableViewCell
            (cell as? SelectParamTableViewCell)?.selectMode = .multi
        case .singleSelectableItem(_):
            cell = tableView.dequeueReusableCell(withIdentifier: "SelectParamTableViewCell", for: indexPath) as! SelectParamTableViewCell
            (cell as? SelectParamTableViewCell)?.selectMode = .single
        default:
            cell = ParamTableViewCell()
            break
        }
        cell.setParamItem(paramItem)
        return cell
    }
    
    public func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        if let paramCell = self.tableView.cellForRow(at: indexPath) as? ParamTableViewCell {
            guard let selectedParamItem = self.paramItems?[indexPath.row] as? ParamItem else { return }
            
            switch selectedParamItem.value {
            case .editableText(_):
                if let cell = (paramCell as? EditableTextParamTableViewCell) {
                    cell.selectItem()
                    cell.updateSwitchUI()
                }
            case .multiSelectableItem(_):
                fallthrough
            case .singleSelectableItem(_):
                //아이템 개별 선택으로 처리.
                break
            default:
                break
            }
        }
        
        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1 * 1) {
           tableView.deselectRow(at: indexPath, animated: false)
        }
    }
    
    public func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return 44
    }
}



